import inspect
from dataclasses import dataclass
from typing import Any, Self

_ERROR_REGISTRY: dict[str, type] = {}


class VeilidAPIError(Exception):
    """Veilid API error exception base class"""

    label = "Base class"

    def __init__(self, *args, **kwargs):
        super().__init__(self.label, *args, **kwargs)

    def __str__(self) -> str:
        args = [('label', self.label)] + sorted(vars(self).items())
        return " ".join(f"{key}={value!r}" for (key, value) in args)

    @classmethod
    def from_json(cls, json: dict) -> Self:
        kind = json["kind"]
        try:
            error_class = _ERROR_REGISTRY[kind]
        except KeyError:
            return cls(f"Unknown exception type: {kind}")

        args = {key: value for key, value in json.items() if key != "kind"}
        return error_class(**args)


@dataclass
class VeilidAPIErrorNotInitialized(VeilidAPIError):
    """Veilid was not initialized"""

    label = "Not initialized"


@dataclass
class VeilidAPIErrorAlreadyInitialized(VeilidAPIError):
    """Veilid was already initialized"""

    label = "Already initialized"


@dataclass
class VeilidAPIErrorTimeout(VeilidAPIError):
    """Veilid operation timed out"""

    label = "Timeout"


@dataclass
class VeilidAPIErrorTryAgain(VeilidAPIError):
    """Operation could not be performed at this time, retry again later"""

    label = "Try again"
    message: str


@dataclass
class VeilidAPIErrorShutdown(VeilidAPIError):
    """Veilid was already shut down"""

    label = "Shutdown"


@dataclass
class VeilidAPIErrorInvalidTarget(VeilidAPIError):
    """Target of operation is not valid"""

    label = "Invalid target"
    message: str


@dataclass
class VeilidAPIErrorNoConnection(VeilidAPIError):
    """Connection could not be established"""

    label = "No connection"
    message: str


@dataclass
class VeilidAPIErrorKeyNotFound(VeilidAPIError):
    """Key was not found"""

    label = "Key not found"
    key: str


@dataclass
class VeilidAPIErrorInternal(VeilidAPIError):
    """Veilid experienced an internal failure"""

    label = "Internal"
    message: str


@dataclass
class VeilidAPIErrorUnimplemented(VeilidAPIError):
    """Functionality is not yet implemented"""

    label = "Unimplemented"
    message: str


@dataclass
class VeilidAPIErrorParseError(VeilidAPIError):
    """Value was not in a parseable format"""

    label = "Parse error"
    message: str
    value: str


@dataclass
class VeilidAPIErrorInvalidArgument(VeilidAPIError):
    """Argument is not valid in this context"""

    label = "Invalid argument"
    context: str
    argument: str
    value: str


@dataclass
class VeilidAPIErrorMissingArgument(VeilidAPIError):
    """Required argument was missing"""

    label = "Missing argument"
    context: str
    argument: str


@dataclass
class VeilidAPIErrorGeneric(VeilidAPIError):
    """Generic error message"""

    label = "Generic"
    message: str


# Build a mapping of canonicalized labels to their exception classes. Do this in-place to update
# the object inside the closure so VeilidAPIError.from_json can access the values.
_ERROR_REGISTRY.clear()
_ERROR_REGISTRY.update(
    {
        obj.label.title().replace(" ", ""): obj
        for obj in vars().values()
        if inspect.isclass(obj) and issubclass(obj, VeilidAPIError)
    }
)


def raise_api_result(api_result: dict) -> Any:
    if "value" in api_result:
        return api_result["value"]
    if "error" in api_result:
        raise VeilidAPIError.from_json(api_result["error"])
    raise ValueError("Invalid format for ApiResult")



from abc import ABC, abstractmethod
from typing import Optional, Self

from . import types
from .state import VeilidState


class RoutingContext(ABC):
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *excinfo):
        if not self.is_done():
            await self.release()

    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    async def release(self):
        pass

    @abstractmethod
    async def with_default_safety(self, release=True) -> Self:
        pass

    @abstractmethod
    async def with_safety(
        self, safety_selection: types.SafetySelection, release=True
    ) -> Self:
        pass

    @abstractmethod
    async def with_sequencing(self, sequencing: types.Sequencing, release=True) -> Self:
        pass

    @abstractmethod
    async def safety(self) -> types.SafetySelection:
        pass

    @abstractmethod
    async def app_call(self, target: types.TypedKey | types.RouteId, request: bytes) -> bytes:
        pass

    @abstractmethod
    async def app_message(self, target: types.TypedKey | types.RouteId, message: bytes):
        pass

    @abstractmethod
    async def create_dht_record(
        self, schema: types.DHTSchema, kind: Optional[types.CryptoKind] = None
    ) -> types.DHTRecordDescriptor:
        pass

    @abstractmethod
    async def open_dht_record(
        self, key: types.TypedKey, writer: Optional[types.KeyPair] = None
    ) -> types.DHTRecordDescriptor:
        pass

    @abstractmethod
    async def close_dht_record(self, key: types.TypedKey):
        pass

    @abstractmethod
    async def delete_dht_record(self, key: types.TypedKey):
        pass

    @abstractmethod
    async def get_dht_value(
        self, key: types.TypedKey, subkey: types.ValueSubkey, force_refresh: bool
    ) -> Optional[types.ValueData]:
        pass

    @abstractmethod
    async def set_dht_value(
        self, key: types.TypedKey, subkey: types.ValueSubkey, data: bytes, writer: Optional[types.KeyPair] = None
    ) -> Optional[types.ValueData]:
        pass

    @abstractmethod
    async def watch_dht_values(
        self,
        key: types.TypedKey,
        subkeys: list[tuple[types.ValueSubkey, types.ValueSubkey]],
        expiration: types.Timestamp,
        count: int,
    ) -> types.Timestamp:
        pass

    @abstractmethod
    async def cancel_dht_watch(
        self,
        key: types.TypedKey,
        subkeys: list[tuple[types.ValueSubkey, types.ValueSubkey]],
    ) -> bool:
        pass

    @abstractmethod
    async def inspect_dht_record(
        self,
        key: types.TypedKey,
        subkeys: list[tuple[types.ValueSubkey, types.ValueSubkey]],
        scope: types.DHTReportScope,
    ) -> types.DHTRecordReport:
        pass



class TableDbTransaction(ABC):
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *excinfo):
        if not self.is_done():
            await self.rollback()

    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    async def commit(self):
        pass

    @abstractmethod
    async def rollback(self):
        pass

    @abstractmethod
    async def store(self, key: bytes, value: bytes, col: int = 0):
        pass

    @abstractmethod
    async def delete(self, key: bytes, col: int = 0):
        pass


class TableDb(ABC):
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *excinfo):
        if not self.is_done():
            await self.release()

    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    async def release(self):
        pass

    @abstractmethod
    async def get_column_count(self) -> int:
        pass

    @abstractmethod
    async def get_keys(self, col: int = 0) -> list[bytes]:
        pass

    @abstractmethod
    async def transact(self) -> TableDbTransaction:
        pass

    @abstractmethod
    async def store(self, key: bytes, value: bytes, col: int = 0):
        pass

    @abstractmethod
    async def load(self, key: bytes, col: int = 0) -> Optional[bytes]:
        pass

    @abstractmethod
    async def delete(self, key: bytes, col: int = 0) -> Optional[bytes]:
        pass


class CryptoSystem(ABC):
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *excinfo):
        if not self.is_done():
            await self.release()

    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    async def release(self):
        pass

    @abstractmethod
    async def cached_dh(self, key: types.PublicKey, secret: types.SecretKey) -> types.SharedSecret:
        pass

    @abstractmethod
    async def compute_dh(
        self, key: types.PublicKey, secret: types.SecretKey
    ) -> types.SharedSecret:
        pass

    @abstractmethod
    async def generate_shared_secret(
        self, key: types.PublicKey, secret: types.SecretKey, domain: bytes
    ) -> types.SharedSecret:
        pass

    @abstractmethod
    async def random_bytes(self, len: int) -> bytes:
        pass

    @abstractmethod
    async def default_salt_length(self) -> int:
        pass

    @abstractmethod
    async def hash_password(self, password: bytes, salt: bytes) -> str:
        pass

    @abstractmethod
    async def verify_password(self, password: bytes, password_hash: str) -> bool:
        pass

    @abstractmethod
    async def derive_shared_secret(self, password: bytes, salt: bytes) -> types.SharedSecret:
        pass

    @abstractmethod
    async def random_nonce(self) -> types.Nonce:
        pass

    @abstractmethod
    async def random_shared_secret(self) -> types.SharedSecret:
        pass

    @abstractmethod
    async def generate_key_pair(self) -> types.KeyPair:
        pass

    @abstractmethod
    async def generate_hash(self, data: bytes) -> types.HashDigest:
        pass

    @abstractmethod
    async def validate_key_pair(self, key: types.PublicKey, secret: types.SecretKey) -> bool:
        pass

    @abstractmethod
    async def validate_hash(self, data: bytes, hash_digest: types.HashDigest) -> bool:
        pass

    @abstractmethod
    async def distance(
        self, key1: types.CryptoKey, key2: types.CryptoKey
    ) -> types.CryptoKeyDistance:
        pass

    @abstractmethod
    async def sign(
        self, key: types.PublicKey, secret: types.SecretKey, data: bytes
    ) -> types.Signature:
        pass

    @abstractmethod
    async def verify(self, key: types.PublicKey, data: bytes, signature: types.Signature):
        pass

    @abstractmethod
    async def aead_overhead(self) -> int:
        pass

    @abstractmethod
    async def decrypt_aead(
        self,
        body: bytes,
        nonce: types.Nonce,
        shared_secret: types.SharedSecret,
        associated_data: Optional[bytes],
    ) -> bytes:
        pass

    @abstractmethod
    async def encrypt_aead(
        self,
        body: bytes,
        nonce: types.Nonce,
        shared_secret: types.SharedSecret,
        associated_data: Optional[bytes],
    ) -> bytes:
        pass

    @abstractmethod
    async def crypt_no_auth(
        self, body: bytes, nonce: types.Nonce, shared_secret: types.SharedSecret
    ) -> bytes:
        pass


class VeilidAPI(ABC):
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *excinfo):
        if not self.is_done():
            await self.release()

    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    async def release(self):
        pass

    @abstractmethod
    async def control(self, args: list[str]) -> str:
        pass

    @abstractmethod
    async def get_state(self) -> VeilidState:
        pass

    @abstractmethod
    async def attach(self):
        pass

    @abstractmethod
    async def detach(self):
        pass

    @abstractmethod
    async def new_private_route(self) -> tuple[types.RouteId, bytes]:
        pass

    @abstractmethod
    async def new_custom_private_route(
        self,
        kinds: list[types.CryptoKind],
        stability: types.Stability,
        sequencing: types.Sequencing,
    ) -> tuple[types.RouteId, bytes]:
        pass

    @abstractmethod
    async def import_remote_private_route(self, blob: bytes) -> types.RouteId:
        pass

    @abstractmethod
    async def release_private_route(self, route_id: types.RouteId):
        pass

    @abstractmethod
    async def app_call_reply(self, call_id: types.OperationId, message: bytes):
        pass

    @abstractmethod
    async def new_routing_context(self) -> RoutingContext:
        pass

    @abstractmethod
    async def open_table_db(self, name: str, column_count: int) -> TableDb:
        pass

    @abstractmethod
    async def delete_table_db(self, name: str) -> bool:
        pass

    @abstractmethod
    async def get_crypto_system(self, kind: types.CryptoKind) -> CryptoSystem:
        pass

    @abstractmethod
    async def best_crypto_system(self) -> CryptoSystem:
        pass

    @abstractmethod
    async def verify_signatures(
        self,
        node_ids: list[types.TypedKey],
        data: bytes,
        signatures: list[types.TypedSignature],
    ) -> list[types.TypedKey]:
        pass

    @abstractmethod
    async def generate_signatures(
        self, data: bytes, key_pairs: list[types.TypedKeyPair]
    ) -> list[types.TypedSignature]:
        pass

    @abstractmethod
    async def generate_key_pair(self, kind: types.CryptoKind) -> list[types.TypedKeyPair]:
        pass

    @abstractmethod
    async def now(self) -> types.Timestamp:
        pass

    @abstractmethod
    async def debug(self, command: str) -> str:
        pass

    @abstractmethod
    async def veilid_version_string(self) -> str:
        pass

    @abstractmethod
    async def veilid_version(self) -> types.VeilidVersion:
        pass

    @abstractmethod
    async def default_veilid_config(self) -> str:
        pass
